@@ -11,8 +11,10 @@ irace.reload.debug <- function(package = "irace")
   options(error=recover)
 }
 
+.irace.prefix <- "== irace == "
+
 .irace.bug.report <-
-  paste("An unexpected condition ocurred.",
+  paste("An unexpected condition occurred.",
         "Please report this bug to the authors of the irace package <http://iridia.ulb.ac.be/irace>")
 
 # Print a user-level fatal error message, when the calling context
@@ -20,7 +22,11 @@ irace.reload.debug <- function(package = "irace")
 ## FIXME: rename this function to irace.error
 tunerError <- function(...)
 {
-  stop (..., call. = FALSE)
+  # The default is only 1000, which is too small. 8170 was the maximum
+  # value allowed in R 2.14.1
+  op <- options(warning.length = 8170)
+  on.exit(options(op))
+  stop (.irace.prefix, ..., call. = FALSE)
 }
 
 irace.assert <- function(exp)
@@ -41,36 +47,36 @@ file.check <- function (file, executable = FALSE, readable = executable,
   READ <- 4
 
   if (!is.character(file) || is.null.or.empty(file)) {
-    stop (text, " ", shQuote(file), " is not a vaild filename")
+    tunerError (text, " ", shQuote(file), " is not a vaild filename")
   }
   ## Remove trailing slash if present for windows OS compatibility
   if (substring(file, nchar(file), nchar(file)) %in% c("/", "\\"))
     file <- substring(file, 1, nchar(file) - 1)
   
   if (!file.exists(file)) {
-    stop (text, " '", file, "' does not exist")
+    tunerError (text, " '", file, "' does not exist")
     return(FALSE)
   }
   if (readable && (file.access(file, mode = READ) != 0)) {
-    stop(text, " '", file, "' is not readable")
+    tunerError(text, " '", file, "' is not readable")
     return (FALSE)
   }
   if (executable && file.access(file, mode = EXEC) != 0) {
-    stop(text, " '", file, "' is not executable")
+    tunerError(text, " '", file, "' is not executable")
     return (FALSE)
   }
 
   if (isdir) {
     if (!file.info(file)$isdir) {
-      stop(text, " '", file, "' is not a directory")
+      tunerError(text, " '", file, "' is not a directory")
       return (FALSE)
     }
     if (notempty && length(list.files (file, recursive=TRUE)) == 0) {
-      stop(text, " '", file, "' does not contain any file")
+      tunerError(text, " '", file, "' does not contain any file")
       return (FALSE)
     }
   } else if (file.info(file)$isdir) {
-    stop(text, " '", file, "' is a directory, not a file")
+    tunerError(text, " '", file, "' is a directory, not a file")
     return (FALSE)
   }
   return (TRUE)
@@ -89,16 +95,136 @@ is.null.or.empty <- function(x)
   is.null(x) || (length(x) == 1 && x == "")
 }
 
-# Function to convert a relative to an absolute path
-path.rel2abs <- function (path)
+strcat <- function(..., collapse = NULL)
+  do.call(paste, args=list(..., sep="", collapse = collapse))
+
+# FIXME: Isn't a R function to do this? More portable?
+canonical.dirname <- function(dirname = stop("required parameter"))
+{
+  return (sub ("([^/])$", "\\1/", dirname))
+}
+
+# Function to convert a relative to an absolute path. CWD is the
+# working directory to complete relative paths. It tries really hard
+# to create canonical paths.
+## FIXME: This needs to be tested on Windows.
+# The following code can be used to test this function.
+## test <- function(testcases) {
+##   cwd <- getwd()
+##   on.exit(setwd(cwd))
+##   setwd("/tmp")
+##   testcases <- read.table(text='
+## "."                          "/tmp"
+## ".."                         "/"
+## "../"                        "/"
+## "../."                       "/"
+## "../.."                     "/"
+## "../../"                     "/"
+## "../../x.r"                  "/x.r"
+## "../leslie/"                 "/leslie"
+## "../leslie/x.r"              "/leslie/x.r"
+## "../x.r"                     "/x.r"
+## "..irace"                    "/tmp/..irace"
+## "./"                         "/tmp"
+## "./."                        "/tmp"
+## "././x.r"                    "/tmp/x.r"
+## "./irace/../x.r"             "/tmp/x.r"
+## "./x.r"                      "/tmp/x.r"
+## ".x.R"                       "/tmp/.x.R"
+## "/./x.r"                     "/x.r"
+## "/home"                      "/home"
+## "/home/leslie/././x.r"       "/home/leslie/x.r"
+## "/home/leslie/~/x.r"         "/home/leslie/~/x.r"
+## "/~/x.r"                     "/~/x.r"
+## "e:/home/leslie/x.r"         "e:/home/leslie/x.r"
+## "leslie/leslie/../../irace"  "/tmp/irace"
+## "x.r"                        "/tmp/x.r"
+## "~/../x.r"                   "/home/x.r"
+## "~/irace/../../x.r"          "/home/x.r"
+## "~/x.r"                      "~/x.r"
+## ', stringsAsFactors=FALSE)
+##   for(i in 1:nrow(testcases)) {
+##     orig <- testcases[i,1]
+##     res <- path.rel2abs(testcases[i,1])
+##     exp <- path.expand(testcases[i,2])
+##     if (res == exp) {
+##       cat("[OK] ", orig, " -> ", res, "\n", sep="")
+##     } else {
+##       stop("[FAILED] ", orig, " -> ", res, " but expected: ", exp, "\n")
+##     }
+##   }
+## }
+## options(error=dump.frames)
+## test(testcases)
+path.rel2abs <- function (path, cwd = getwd())
 {
   if (is.null.or.na(path)) {
     return (NULL)
   } else if (path == "") {
     return ("")
-  } else {
-    return (sub ('^(\\.)', paste (getwd(), '/\\1', sep=''), path))
   }
+
+  # Possibly expand ~/path to /home/user/path.
+  path <- path.expand(path)
+
+  filename <- basename(path)
+  path <- dirname(path)
+  s <- .Platform$file.sep
+  if (path == ".") {
+    if (filename == ".") { # This is the current directory
+      return (cwd)
+      # We handle the case ".." later.
+    } else if (filename != "..") { # This is a file in the current directory
+      return(strcat(cwd, s, filename))
+    }
+  }
+  # else
+  
+  # Prefix the current cwd to the path if it doesn't start with "c:\" or /
+  reg.exp <- strcat("^", s, "|^[A-Za-z]:", s)
+  if (!grepl(reg.exp, path))
+    path <- strcat(cwd, s, path)
+
+  # Change "/./" to "/" to get a canonical form 
+  path <- gsub(strcat(s, ".", s), s, path, fixed = TRUE)
+
+  # Change "/.$" to "/" to get a canonical form 
+  path <- sub(strcat(s, "\\.$"), s, path)
+
+  # Change "/x/../" to "/" to get a canonical form 
+  prevdir.regex <- strcat(s, "[^", s,"]+", s, "\\.\\.")
+  repeat {
+    # We need to do it one by one so "a/b/c/../../../" is not converted to "a/b/../"
+    tmp <- sub(strcat(prevdir.regex, s), s, path)
+    if (tmp == path) break
+    path <- tmp
+  }
+  # Handle "/something/..$" to "/" that is, when ".." is the last thing in the path.
+  path <- sub(strcat(prevdir.regex, "$"), s, path)
+
+  # Handle "^/../../.." to "/" that is, going up at the root just returns the root.
+  repeat {
+    # We need to do it one by one so "a/b/c/../../../" is not converted to "a/b/../"
+    tmp <- sub(strcat("^", s, "\\.\\.", s), s, path)
+    if (tmp == path) break
+    path <- tmp
+  }
+  # Handle "^/..$" to "/" that is, when ".." is the last thing in the path.
+  path <- sub(strcat("^", s, "\\.\\.$"), s, path)
+
+  # It may happen that path ends in "/", for example, for "/x". Do
+  # not add another "/"
+  if (filename != ".") {
+    last <- substr(path, nchar(path), nchar(path))
+    if (last == s) {
+      path <- strcat(path, filename)
+    } else {
+      path <- strcat(path, s, filename)
+    }
+  }
+  # We use normalizePath, which will further simplify the path if
+  # the path exists.
+  return (suppressWarnings(normalizePath(path, mustWork = NA)))
 }
 
 is.function.name <- function(FUN)
@@ -111,16 +237,6 @@ is.function.name <- function(FUN)
                  inherits = TRUE)[[1]]))
 }
 
-# FIXME: Isn't a R function to do this? More portable?
-canonical.dirname <- function(dirname = stop("required parameter"))
-{
-  return (sub ("([^/])$", "\\1/", dirname))
-}
-
-# FIXME: This is paste0 in R >= 2.15
-strcat <- function(..., sep = "", collapse = NULL)
-  paste(list(...), sep, collapse)
-
 trim.leading <- function(str)
 {
   return (sub('^[[:space:]]+', '', str)) ## white space, POSIX-style
@@ -217,14 +333,19 @@ mpiInit <- function(nslaves, debugLevel = 0)
   # Load the Rmpi package if it is not already loaded.
   if (!is.loaded("mpi_initialize")) {
     if (! require("Rmpi", quietly = TRUE))
-      stop("The `Rmpi' package is required for using MPI.")
+      tunerError("The `Rmpi' package is required for using MPI.")
 
+    # FIXME: We should do this when irace finalizes.
     # When R exits, finalize MPI.
     reg.finalizer(environment(Rmpi::mpi.exit), function(e) {
       # Rmpi already prints a message, so we don't need this.
       # cat("# Finalize MPI...\n")
       if (Rmpi::mpi.comm.size(1) > 0)
-        Rmpi::mpi.close.Rslaves()
+        # FIXME: dellog == TRUE tries to delete log files, but it does
+        # not take into account that we may have change directory and
+        # it does not fails gracefully but produces an annoying:
+        # Warning message: running command 'ls *.30577+1.*.log 2>/dev/null' had status 2
+        Rmpi::mpi.close.Rslaves(dellog = FALSE)
       # FIXME: How to avoid the message
       # "Rmpi cannot be used unless relaunching R" ?
       Rmpi::mpi.exit()
